'''
Tests for the algo.multicorr module.
'''

import unittest
import numpy as np
import matplotlib.pyplot as plt
import os
import sys
from io import StringIO

import ncempy.algo.multicorr as mc

class test_shiftedimages(unittest.TestCase):
    """Test the multicorr module on two shifted images"""

    def test_parse_input(self):
        '''
        Test to make sure inputs are properly accounted for.
        '''


        g1 = 1
        g2 = np.zeros([4,4])
        with self.assertRaises(TypeError):
            mc.parse_input(g1, g2) # both must be arrays
        with self.assertRaises(TypeError):
            mc.parse_input(g2, g1)

        # tests input sanitation.
        saved_stdout = sys.stdout
        try:
            out = StringIO()
            sys.stdout = out
            mc.parse_input(g2, g2, 'dummy', 4)
            output = out.getvalue().strip()
            self.assertEqual(output, 'Unknown method used, setting to cross')

            out = StringIO()
            sys.stdout = out
            mc.parse_input(g2, g2, 'cross', 23.1)
            output = out.getvalue().strip()
            self.assertEqual(output, 'Upsample factor is not an integer, rounding down')

            out = StringIO()
            sys.stdout = out
            mc.parse_input(g2, g2, 'cross', -23.1)
            output = out.getvalue().strip()
            self.assertEqual(output, 'Upsample factor is not an integer, rounding down\nUpsample factor is < 1, setting to 1')

            out = StringIO()
            sys.stdout = out
            mc.parse_input(g2, g2, 'cross', 'phase')
            output = out.getvalue().strip()
            self.assertEqual(output, 'Upsample factor is not an integer or float, setting to 1')
        finally:
            sys.stdout = saved_stdout

        with self.assertRaises(TypeError):
            mc.parse_input(np.zeros([2,3]), np.zeros([3,2]), 'phase', 1)

        with self.assertRaises(TypeError):
            mc.initial_correlation_image(g2, g2, 'tester', 1)

    def test_multicorr(self):
        '''
        Test to check if the correlation is working.
        '''
        g2 = np.zeros((3,3))
        test = mc.multicorr(np.fft.fft2(g2), np.fft.fft2(g2), 'phase', 3)
        print(test)
        assert list(test) == [0, 0]

        # filename = '/Users/Tom/Downloads/matt_beard.jpg'
        # filename_shifted = '/Users/Tom/Downloads/matt_beard_shifted.jpg'
        # G1 = cv2.imread(filename, 0)
        # G1 = G1[0:100, 0:100]
        G1 = np.zeros((100,100))
        G1[42,35] = 12
        # G2 = cv2.imread(filename_shifted, 0)
        G2 = np.real(np.fft.ifft2(mc.imageShifter(np.fft.fft2(G1), [-30, 10])))
        print(G1.shape, G2.shape)
        # plt.figure(0)
        # plt.imshow(np.concatenate((G1, G2)))
        # plt.show(block = True)
        imageCorr = mc.initial_correlation_image(np.fft.fft2(G1), np.fft.fft2(G2), 'phase', 2)

        # test to see if correlation image looks the same
        np.testing.assert_almost_equal(np.exp(1j * np.angle((np.multiply(np.fft.fft2(G2), np.conj(np.fft.fft2(G1)))))), imageCorr, decimal=4)

        # plt.figure(1)
        # plt.imshow(np.real(np.fft.ifft2(imageCorr)))
        # plt.show(block = True )
        out_phase = mc.multicorr(np.fft.fft2(G1), (np.fft.fft2(G2)), 'phase', 2)
        self.assertEqual(out_phase, [-30.0, 10.0])

        # test cross correlation
        out_cross = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'cross', 1)
        self.assertEqual(out_cross, [-30.0, 10.0])

        # test hybrid correlation and dft upsample
        out_hybrid = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'hybrid', 3)
        out_hybrid = list(out_hybrid)
        np.testing.assert_almost_equal(out_hybrid, [-30.0, 10.0], decimal = 6)

        # plt.imshow(np.subtract(G1, np.real(np.fft.ifft2(out.G2shift))))
        # plt.show(block = True)


    def test_different_shifts(self):
        '''
        Tests to check if mod is working correctly
        '''
        # filename = '/Users/Tom/Downloads/matt_beard.jpg'
        # G1 = cv2.imread(filename, 0)
        # G1 = G1[0:501, 0:501]
        G1 = np.zeros((101,101))
        G1[55,55] = 12
        # plt.imshow(G1)
        # plt.show(block = True)
        shifts = [[3., 1.], [-3., -1.], [-3., 1.], [3.,-1.]]
        shifts2 = [[10.3, 14.1], [-10.3, -14.1], [-10.3, 14.1], [10.3,-14.1]]
        for i in shifts:
            # print(i)
            G2 = np.real(np.fft.ifft2(mc.imageShifter(np.fft.fft2(G1), i)))
            # plt.imshow(np.concatenate((G1, G2)))
            # plt.show(block = True)
            with self.subTest(i = i):
                out_phase = mc.multicorr(np.fft.fft2(G1), (np.fft.fft2(G2)), 'phase', 3)
                np.testing.assert_almost_equal(out_phase, i, decimal = 6)
                out_cross = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'cross', 3)
                np.testing.assert_almost_equal(out_cross, i, decimal = 6)
                out_hybrid = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'hybrid', 3)
                out_hybrid = list(out_hybrid)
                np.testing.assert_almost_equal(out_hybrid, i, decimal = 6)
        for i in shifts2:
            # print(i)
            G2 = np.real(np.fft.ifft2(mc.imageShifter(np.fft.fft2(G1), i)))
            with self.subTest(i = i):

                out_phase = mc.multicorr(np.fft.fft2(G1), (np.fft.fft2(G2)), 'phase', 10)
                out_phase = list(out_phase)
                np.testing.assert_almost_equal(out_phase, i, decimal = 2)

                out_hybrid = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'hybrid', 10)
                out_hybrid = list(out_hybrid)
                np.testing.assert_almost_equal(out_hybrid, i, decimal = 2)

                out_cross = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'cross', 10)
                out_cross = list(out_cross)
                np.testing.assert_almost_equal(out_cross, i, decimal = 2)

if __name__ == "__main__":
    unittest.main()



'''
Module to correlate two images, functionally written.

TODO
----
    - Replace makeFourierCoords with np.fft.fftfreq
    - Add verbose output
    - Add testing for complex input for G1 and G2
    - upsampled_correlation has ```if upSample > 2``` but
      should it be if upSample >= 2?
    - Cant use rfft2 currently. This gives one shift as 1/2 the value. How
      can this be improved to improve speed?
    - Better to make this a class. Then you can set it
      up and change the inputs without having to run
      through all of the setup again
    - I think upSampleFFT can be replaced by np.fft.fft2
      with the s option.
    - imageShifter and multicorr output have opposite sign.

'''

import numpy as np

def multicorr(G1, G2, method = 'cross', upsampleFactor = 1):
    '''Align a template to an image by cross correlation. THe template
    and the image must have the same size.
    
    The function takes in FFTs so that any FFT algorithm can be used to 
    transform the image and template (fft2, mkl, scipack, etc.)
    
    Parameters   
    ----------
        G1 : complex ndarray
            Fourier transform of reference image.
        G2 : complex ndarray
            Fourier transform of the image to register (the kernel).
        method : str, optional
            The correlation method to use. Must be 'phase' or 'cross' or 'hybrid' (default = 'cross')
        upsampleFactor : int
            Upsample factor for subpixel precision of cross correlation. (default = 1)

    Returns
    -------
        xyShift : list of floats
            The shift between G1 and G2 in pixels.
    
    Example
    -------
        Cross correlate two images already stored as ndarrays. You must input the FFT
        of the images.
        
        >>> import ncempy.multicorr as mc
        >>> import numpy as np
        >>> im0FFT = np.fft.fft2(im0)
        >>> im1FFT = np.fft.fft2(im1)
        >>> shifts = mc.multicorr(im0FFT,im1FFT)
    
    '''
    
    #method, upsampleFactor = parse_input(G1, G2, method, upsampleFactor)
    # Check to make sure both G1 and G2 are arrays
    if type(G1) is not np.ndarray:
        raise TypeError('G1 must be an ndarray')
    elif type(G2) is not np.ndarray:
        raise TypeError('G2 must be an ndarray')
    
    #Check that the inputs are complex FFTs (common error)
    if not np.iscomplexobj(G1) or not np.iscomplexobj(G2):
        raise TypeError('G1 and G2 must be complex FFTs.')
    
    # Check to make sure method and upsample factor are the correct values
    if method not in ['phase', 'cross', 'hybrid']:
        print('Unknown method used, setting to cross.')
        method = 'cross'

    if type(upsampleFactor) is not int and type(upsampleFactor) is not float:
        print('Upsample factor is not an integer or float, setting to 1')
        upsampleFactor = 1
    elif type(upsampleFactor) is not int:
        print('Upsample factor is not an integer, rounding down')
        upsampleFactor = int(upsampleFactor)
        if upsampleFactor < 1:
            print('Upsample factor is < 1, setting to 1')
            upsampleFactor = 1

    # Verify images are the same size.
    if G1.shape != G2.shape:
        raise TypeError('G1 and G2 are not the same size, G1 is {0} and G2 is {1}'.format(G1.shape, G2.shape))

    
    imageCorr = initial_correlation_image(G1, G2, method, upsampleFactor)
    xyShift = upsampled_correlation(imageCorr, upsampleFactor)
    
    return xyShift

def initial_correlation_image(G1, G2, method = 'cross', upsampleFactor = 1):
    '''Generate correlation image at initial resolution using the method specified.

    Parameters   
    ----------
        G1 : complex ndarray
            Fourier transform of reference image.
        G2 : complex ndarray
            Fourier transform of the image to register (the kernel).
        method : str, optional
            The correlation method to use. Must be 'phase' or 'cross' or 'hybrid' (default = 'cross')
        upsampleFactor : int
            Upsample factor for subpixel precision of cross correlation. (default = 1)

    Returns
    -------
        imageCorr : ndarray complex
            Correlation array which has not yet been inverse Fourier transformed.
    '''
    G12 = np.multiply(G2, np.conj(G1)) # is this the correct order that we want?
    if method == 'phase':
        imageCorr = np.exp(1j * np.angle(G12))
    elif method == 'cross':
        imageCorr = G12
    elif method == 'hybrid':
        imageCorr = np.multiply(np.sqrt(np.absolute(G12)), np.exp(1j * np.angle(G12)))
    else:
        raise TypeError('{} method is not allowed'.format(str(method)))

    return imageCorr

def upsampled_correlation(imageCorr, upsampleFactor):
    '''Upsamples the correlation image by a set integer factor upsampleFactor.
    If upsampleFactor == 2, then it is naively Fourier upsampled.
    If the upsampleFactoris higher than 2, then it uses dftUpsample, which is
    a more efficient way to Fourier upsample the image.

    Parameters
    ----------
        imageCorr : ndarray complex 
            Fourier transformed correlation image returned by initial_correlation_image.
        upsampleFactor : int
            Upsampling factor.

    Returns
    -------
        xyShift : list
            Shift in x and y of G2 with respect to G1.
    '''

    imageCorrIFT = np.real(np.fft.ifft2(imageCorr))
    xyShift = list(np.unravel_index(imageCorrIFT.argmax(), imageCorrIFT.shape, 'C'))

    if upsampleFactor == 1:
        imageSize = imageCorrIFT.shape
        xyShift[0] = ((xyShift[0] + imageSize[0]/2) % imageSize[0]) - imageSize[0]/2
        xyShift[1] = ((xyShift[1] + imageSize[1]/2) % imageSize[1]) - imageSize[1]/2
        
    else:
        imageCorrLarge = upsampleFFT(imageCorr, 2)
        imageSizeLarge = imageCorrLarge.shape
        xySubShift2 = list(np.unravel_index(imageCorrLarge.argmax(), imageSizeLarge, 'C'))
        #print('xySubShift2 = {}'.format(xySubShift2))
        xySubShift2[0] = ((xySubShift2[0] + imageSizeLarge[0]/2) % imageSizeLarge[0]) - imageSizeLarge[0]/2
        xySubShift2[1] = ((xySubShift2[1] + imageSizeLarge[1]/2) % imageSizeLarge[1]) - imageSizeLarge[1]/2
        xyShift = [i/2 for i in xySubShift2] #signs have to flip, or mod wrong?
        #print('xyShiftln127 = {}.format(xyShift))

        if upsampleFactor > 2:
            # here is where we use DFT registration to make things much faster
            # we cut out and upsample a peak 1.5 by 1.5 px from our original correlation image.

            xyShift[0] = np.round(xyShift[0] * upsampleFactor) / upsampleFactor
            xyShift[1] = np.round(xyShift[1] * upsampleFactor) / upsampleFactor

            globalShift = np.fix(np.ceil(upsampleFactor * 1.5)/2)# this line might have an off by one error based. The associated matlab comment is "this will be used to center the output array at dftshift + 1"
            #print('globalShift', globalShift, 'upsampleFactor', upsampleFactor, 'xyShift', xyShift)

            imageCorrUpsample = np.conj(dftUpsample(np.conj(imageCorr), upsampleFactor, globalShift - np.multiply(xyShift, upsampleFactor))) / (np.fix(imageSizeLarge[0]) * np.fix(imageSizeLarge[1]) * upsampleFactor ** 2)

            xySubShift = np.unravel_index(imageCorrUpsample.argmax(), imageCorrUpsample.shape, 'C')
            #print('xySubShift = {}'.format(xySubShift))

            # add a subpixel shift via parabolic fitting
            try:
                icc = np.real(imageCorrUpsample[xySubShift[0] - 1 : xySubShift[0] + 2, xySubShift[1] - 1 : xySubShift[1] + 2])
                dx = (icc[2,1] - icc[0,1]) / (4 * icc[1,1] - 2 * icc[2,1] - 2 * icc[0,1])
                dy = (icc[1,2] - icc[1,0]) / (4 * icc[1,1] - 2 * icc[1,2] - 2 * icc[1,0])
            except:
                dx, dy = 0, 0 # this is the case when the peak is near the edge and one of the above values does not exist
            #print('dxdy = {}, {}'.format(dx, dy))
            #print('xyShift = {}'.format(xyShift))
            xySubShift = xySubShift - globalShift;
            #print('xysubShift2 = {}'.format(xySubShift))
            xyShift = xyShift + (xySubShift + np.array([dx, dy])) / upsampleFactor
            #print('xyShift2 = {}'.format(xyShift))

    return xyShift

def upsampleFFT(imageInit, upsampleFactor):
    '''This does a Fourier upsample of the imageInit. imageInit is the Fourier transform of the correlation image. 
    The function returns the real space correlation image that has been Fourier upsampled by the upsampleFactor.
    An upsample factor of 2 is generally sufficient.

    The way it works is that it embeds imageInit in a larger array of zeros, then does the inverse Fourier transform to return the Fourier upsampled image in real space.

    Parameters
    ----------
        imageInit : ndarray  complex
            The image to be Fourier upsampled. This should be in the Fourier domain.
        upsampleFactor : int
            THe upsample factor (usually 2).

    Returns
    -------
        imageUpsampleReal : ndarray complex 
            The inverse Fourier transform of imageInit upsampled by the upsampleFactor.
            
    TODO
    ----
        This can be fully replaced by np.fft.ffts(im,s=(1024,1024)) where im.shape = (512,512)
        Also change to use rfft2
    '''
    imageSize = imageInit.shape
    imageUpsample = np.zeros(tuple((i*upsampleFactor for i in imageSize))) + 0j
    imageUpsample[:imageSize[0], :imageSize[1]] = imageInit
    imageUpsample = np.roll(np.roll(imageUpsample, -int(imageSize[0]/2), 0), -int(imageSize[1]/2),1)
    imageUpsampleReal = np.real(np.fft.ifft2(imageUpsample))
    return imageUpsampleReal

def dftUpsample(imageCorr, upsampleFactor, xyShift):
    '''
    This performs a matrix multiply DFT around a small neighboring region of the inital correlation peak.
    By using the matrix multiply DFT to do the Fourier upsampling, the efficiency is greatly improved.
    This is adapted from the subfuction dftups found in the dftregistration function on the Matlab File Exchange.

    https://www.mathworks.com/matlabcentral/fileexchange/18401-efficient-subpixel-image-registration-by-cross-correlation

    The matrix multiplication DFT is from 
    Manuel Guizar-Sicairos, Samuel T. Thurman, and James R. Fienup, "Efficient subpixel image registration algorithms," 
    Opt. Lett. 33, 156-158 (2008). http://www.sciencedirect.com/science/article/pii/S0045790612000778

    Parameters
    ----------
        imageCorr : ndarray
            Correlation image between two images in Fourier space.
        upsampleFactor : int 
            Scalar integer of how much to upsample.
        xyShift : list of 2 floats 
            Single pixel shift between images previously computed. Used to center the matrix multiplication 
            on the correlation peak.
        
    Returns
    -------
        imageUpsample : ndarray
            Upsampled image from region around correlation peak.
        
    '''
    imageSize = imageCorr.shape
    pixelRadius = 1.5
    numRow = np.ceil(pixelRadius * upsampleFactor)
    numCol = numRow

    colKern = np.exp(
    (-1j * 2 * np.pi / (imageSize[1] * upsampleFactor))
    * (np.fft.ifftshift( (np.arange(imageSize[1])) )
    - np.floor(imageSize[1]/2))
    * (np.arange(numCol) - xyShift[1])[:, np.newaxis]
    ) # I think this can be written differently without the need for np.newaxis. This might require np.outer to compute the matrix itself instead of just using np.dot.

    rowKern = np.exp(
    (-1j * 2 * np.pi / (imageSize[0] * upsampleFactor))
    * (np.arange(numRow) - xyShift[0])
    * (np.fft.ifftshift(np.arange(imageSize[0]))
    - np.floor(imageSize[0]/2))[:, np.newaxis]
    ) # Comment from above applies.

    imageUpsample = np.real(np.dot(np.dot(rowKern.transpose(), imageCorr), colKern.transpose()))

    return imageUpsample

def imageShifter(G2, xyShift):
    '''
    This function multiplies G2 by a plane wave that has the real space effect of shifting ifft2(G2) by [x, y] pixels.

    Parameters
   -----------
        G2 : complex ndarray
            The Fourier transform of an image.
        xyShift : list
            A two element list of the shifts along each axis.

    Returns
    -------
        G2shift : complex ndarray
            Fourier shifted image FFT
            
    Example
    -------
    
        >>> shiftIm0 = np.real(np.fft.ifft2(multicorr.imageShifter(np.fft.fft2(im0),[11.1,22.2])))
        >>> plt.imshow(shiftIm0)
        
    '''
    imageSize = G2.shape
    qx = makeFourierCoords(imageSize[0], 1) # does this need to be a column vector
    if imageSize[1] == imageSize[0]:
        qy = qx
    else:
        qy = makeFourierCoords(imageSize[1], 1)

    G2shift = np.multiply(G2, np.outer( np.exp(-2j * np.pi * qx * xyShift[0]),  np.exp(-2j * np.pi * qy * xyShift[1])))

    return G2shift

def makeFourierCoords(N, pSize):
    '''
    This function creates Fourier coordinates such that (0,0) is in the center of the array.

    Parameters
    ----------
        N :int
            The maximum coordinate in the original frame.
        pSize : float
            The pixel size.

    Returns
    -------
        q : ndarray
            A single row array that has transformed 0:N to -N/2:N/2, such that the array sizes are the same.
    '''

    N = float(N)
    if N % 2 == 0:
        q = np.roll(np.arange(-N/2, N/2, dtype = 'float64') / (N * pSize), int(-N/2), axis=0)
    else:
        q = np.roll(np.arange((1-N)/2, (N+1)/2) / (N * pSize), int((1-N)/2), axis=0)
    return q



''' Provides an interface to Velox EMD datasets. Not to be confused with
Berkeley EMD data sets (see emd.py) instead.

The reader for EMD Berkeley and Velox files will be combined in the near
future once they are fully tested separately.

Currently limited to only images. This file can not load spectra.
'''

import json

import numpy as np
import h5py

import h5py_cache

class fileEMDVelox:
    '''Class to represent Velox EMD files. It uses the h5py_cache module
    to increase the default cache size from 1MB to 10MB. This significantly
    improves file reading for EMDVelox files which are written with Fortran-
    style ordering and a poor choice of chunking.
    
    Parameters
    ----------
        filename : str
            Name of the EMD file.
    
    Attributes
    ----------
        list_data : list
            A list containing each h5py data group.
        
        file_hdl : h5py.File
            The File handle from h5py.File.
        
        metaDataJSON : dict
            The metadata for the most recently loaded data set.
        
    
    Example
    -------
        Open an EMD Velox file of 1 image.
        
        >>> import ncempy.io as nio
        >>> emd1 = nio.emdVelox.fileEMDVelox('1435 1.2 Mx STEM HAADF-DF4-DF2-BF.emd')
        >>> print(emd1) # print informatio about the file
        >>> im0, metadata0 = emd1.get_dataset(emd1.list_data[0])
        >>> plt.imshow(im0, extent = (0,md1['pixelSize'][0],0,md1['pixelSize'][1]))
    '''
    
    def __init__(self, filename):
        '''Init opening the file and finding all data groups. Currently only
        searches the /Data/Images group.
        
        Parameters
        ----------
            filename :
                The file path to load as a string.
        
        '''
        
        ## necessary declaration in case something goes wrong
        self.file_hdl = None
        self.metaDataJSON = None
        self.list_data = None
        
        # check for string
        if not isinstance(filename, str):
            raise TypeError('Filename is supposed to be a string!')

        # try opening the file
        try:
            #self.file_hdl = h5py.File(filename, 'r')
            self.file_hdl = h5py_cache.File(filename, 'r', chunk_cache_mem_size=10*1024**2)
        except:
            print('Error opening file for readonly: "{}"'.format(filename))
            raise
            
        self._find_groups()
        
    def __del__(self):
        '''Destructor for EMD file object.
        
        Closes the h5py file.
        
        '''
        # close the file
        self.file_hdl.close()

    def __enter__(self):
        '''Implement python's with staement
        
        '''
        return self
        
    def __exit__(self,type,value,traceback):
        '''Implement python's with statment
        and close the file via __del__()
        
        '''
        self.__del__()
        return None
    
    def __str__(self):
        ''' Print out the detectors used to take the data and
        the pixel size to help with telling users about the data in the file.
        
        '''
        out = 'EMD file contains {} data sets\n'.format(len(self.list_data))
        for ii, group in enumerate(self.list_data):
            md = self.parseMetaData(group)
            out += 'Dataset #{} from detector: {}\n'.format(ii,md['detectorName'])
        out += 'pixel size = ({0[0]:0.4f}, {0[1]:0.4f}) nm'.format(md['pixelSize'])
        return out
    
    def _find_groups(self):
        '''Find all groups that contain data.
        
        Note
        ----
            This currently only finds images.
        '''
        try:
            # Get all of the groups in the Image group
            self.list_data = list(self.file_hdl['Data/Image'].values())
        except:
            self.list_data = []
            raise
    
    def get_dataset(self, group):
        '''Get the data from a group and the associated metadata.
        
        Parameters
        ----------
            group : HDF5 dataset or int
                The link to the HDF5 dataset in the file or an integer for the 
                number of the dataset. The list of datasets is held in the
                list_data attribute populated on class init.
        '''
        # check input
        try:
            if type(group) is int:
                group = self.list_data[group]
        except IndexError:
            raise IndexError('EMDVelox group #{} does not exist.'.format(group))
        
        if not isinstance(group, h5py._hl.group.Group):
            raise TypeError('group needs to refer to a valid HDF5 group!')

        data = np.squeeze(group['Data'][:]) #load the full data set
        metaData = self.parseMetaData(group)
        return (data,metaData)
    
    def parseMetaData(self,group):
        '''Parse metadata in a data group. Determines the pixelSize and 
        detector name. The EMDVelox data sets have extensive metadata
        stored as a JSON type string.
        
        Parameters
        ----------
            group : h5py group
                The h5py group to load the metadata from. The string is loaded
                and parsed by the json module into a dictionary
        
        Returns
        -------
            md : dict
                The JSON information returned as a python dictionary.
        
        '''
        md = {}
        tempMetaData = group['Metadata'][:,0]
        # Reduce to valid metadata
        validMetaDataIndex = np.where(tempMetaData > 0) 
        metaData = tempMetaData[validMetaDataIndex].tostring()
        # Interpret as UTF-8 encoded characters and load as JSON
        self.metaDataJSON = json.loads(metaData.decode('utf-8','ignore'))
        # Pull out basic meta data about the images
        detectorName = self.metaDataJSON['BinaryResult']['Detector']
        pixelSizeX = float(self.metaDataJSON['BinaryResult']['PixelSize']['width'])*1e9 #change to nm
        pixelSizeY = float(self.metaDataJSON['BinaryResult']['PixelSize']['height'])*1e9 #change to nm
        # Construct meta data dictionary
        md['pixelSize'] = (pixelSizeX,pixelSizeY)
        md['pixelSizeUnit'] = ('nm','nm')
        md['detectorName'] = detectorName
        
        return md
    